@@ -154,14 +154,36 @@ class SourceDocMongoBackend(SourceDocBackendBase):
             # are sub-source names, not main source so we have to deal with src_dump as well
             # in order to resolve/map main/sub source name
             subsrc_versions = []
+
+            if src and src.get("download"):
+                # Store the latest success dump time
+                src_meta.setdefault(src["_id"], {})
+                last_success = src["download"].get("last_success")
+                if not last_success and src["download"].get("status") == "success":
+                    last_success = src["download"].get("started_at")
+                if last_success:
+                    src_meta[src["_id"]]["download_date"] = last_success
+
             if src and src.get("upload"):
+                latest_upload_date = None
                 meta = {}
                 for job_name in src["upload"].get("jobs", {}):
                     job = src["upload"]["jobs"][job_name]
                     # "step" is the actual sub-source name
-                    docm = self.master.find_one({"_id": job.get("step")})
+                    sub_source = job.get("step")
+                    docm = self.master.find_one({"_id": sub_source})
                     if docm and docm.get("src_meta"):
-                        meta[job.get("step")] = docm["src_meta"]
+                        meta[sub_source] = docm["src_meta"]
+                    # Store the latest success upload time
+                    if not latest_upload_date or latest_upload_date < job["started_at"]:
+                        step_meta = meta.setdefault(sub_source, {})
+                        sub_source_info = src["upload"]["jobs"][sub_source]
+                        last_success = sub_source_info.get("last_success")
+                        if not last_success and sub_source_info.get("status") == "success":
+                            last_success = sub_source_info.get("started_at")
+                        if last_success:
+                            step_meta["upload_date"] = last_success
+
                 # when more than 1 sub-sources, we can have different version in sub-sources
                 # (not normal) if one sub-source uploaded, then dumper downloaded a new version,
                 # then the other sub-source uploaded that version. This should never happen, just make sure
@@ -318,7 +340,7 @@ def create_backend(db_col_names, name_only=False, follow_ref=False, **kwargs):
         col = db[db_col_names[1]]
         # normalize params (0:host, 1:port)
         db_col_names = [
-            "%s:%s" % (db.client.address[0], db.client.address[1]), db.name,
+            "%s:%s" % (db.client.HOST, db.client.PORT), db.name,
             col.name
         ]
     assert col is not None, "Could not create collection object from %s" % repr(
@@ -359,7 +381,7 @@ def merge_src_build_metadata(build_docs):
         "More than one build document must be passed in order metadata"
     first = build_docs[0]
     others = build_docs[1:]
-    meta = first.get("_meta", {})
+    meta = first.setdefault("_meta", {})
     for new_doc in others:
         new_meta = new_doc.get("_meta", {})
         for k, v in new_meta.items():
