@@ -10,6 +10,7 @@ import random
 
 from biothings.utils.common import timesofar, get_timestamp, \
     dump, rmdashfr, loadobj, md5sum
+from biothings.utils.serializer import to_json, to_json_file
 from biothings.utils.mongo import id_feeder, get_target_db, get_previous_collection
 from biothings.utils.hub_db import get_src_build
 from biothings.utils.loggers import get_logger
@@ -297,7 +298,7 @@ class BaseDiffer(object):
                             got_error = err
                     self.metadata["diff"]["mapping_file"] = res["mapping_file"]
                     diff_stats["mapping_changed"] = True
-                    json.dump(self.metadata,
+                    to_json_file(self.metadata,
                               open(self.metadata_filename, "w"),
                               indent=True)
 
@@ -394,7 +395,7 @@ class BaseDiffer(object):
             self.logger.info(
                 "Finished calculating diff for the old collection. Total number of docs deleted: {}"
                 .format(diff_stats["delete"]))
-            json.dump(self.metadata,
+            to_json_file(self.metadata,
                       open(self.metadata_filename, "w"),
                       indent=True)
 
@@ -478,7 +479,7 @@ class BaseDiffer(object):
                                  job={"step": "diff-reduce"})
             res = await merge_diff()
             self.metadata["diff"]["files"] = res
-            json.dump(self.metadata,
+            to_json_file(self.metadata,
                       open(self.metadata_filename, "w"),
                       indent=True)
             if got_error:
@@ -520,7 +521,7 @@ class BaseDiffer(object):
 
             job.add_done_callback(posted)
             await job
-            json.dump(self.metadata,
+            to_json_file(self.metadata,
                       open(self.metadata_filename, "w"),
                       indent=True)
             if got_error:
@@ -1322,7 +1323,7 @@ class DifferManager(BaseManager):
                     metadata["diff"]["mapping"] = info
                 else:
                     metadata["diff"]["files"].append(info)
-            json.dump(metadata,
+            to_json_file(metadata,
                       open(os.path.join(diff_folder, "metadata.json"), "w"),
                       indent=True)
             self.logger.info(
