@@ -1,7 +1,8 @@
-import asyncio
+import os
+import subprocess
 
-from biothings.utils.hub_db import get_data_plugin
 import biothings.hub.dataload.dumper as dumper
+from biothings.utils.hub_db import get_data_plugin
 
 
 class GitDataPlugin(dumper.GitDumper):
@@ -10,7 +11,7 @@ class GitDataPlugin(dumper.GitDumper):
     # so we don't mix data sources and plugins
     def prepare_src_dump(self):
         self.src_dump = get_data_plugin()
-        self.src_doc = self.src_dump.find_one({'_id': self.src_name}) or {}
+        self.src_doc = self.src_dump.find_one({"_id": self.src_name}) or {}
 
 
 class ManualDataPlugin(dumper.ManualDumper):
@@ -19,17 +20,18 @@ class ManualDataPlugin(dumper.ManualDumper):
     # so we don't mix data sources and plugins
     def prepare_src_dump(self):
         self.src_dump = get_data_plugin()
-        self.src_doc = self.src_dump.find_one({'_id': self.src_name}) or {}
+        self.src_doc = self.src_dump.find_one({"_id": self.src_name}) or {}
 
-    @asyncio.coroutine
-    def dump(self, *args, **kwargs):
-        yield from super(ManualDataPlugin, self).dump(
+    async def dump(self, *args, **kwargs):
+        await super(ManualDataPlugin, self).dump(
             path="",  # it's the version is original method implemention
             # but no version here available
-            release="", *args, **kwargs)
+            release="",
+            *args,
+            **kwargs,
+        )
 
 
 class DataPluginManager(dumper.DumperManager):
-
     def load(self, plugin_name, *args, **kwargs):
         return super(DataPluginManager, self).dump_src(plugin_name, *args, **kwargs)
