@@ -3,30 +3,32 @@
 # To run this program, the file ``ssh_host_key`` must exist with an SSH
 # private key in it to use as a server host key.
 
-import sys
-import os
-import re
-import datetime
-import json
-import pydoc
-import copy
 import asyncio
+import copy
+import datetime
 import io
+import json
+import os
+import pydoc
+import re
+import sys
 import time
+from collections import OrderedDict, UserDict
 from functools import partial
 from pprint import pformat
-from collections import OrderedDict
 
 from IPython import InteractiveShell
 
-from biothings.utils.dataload import to_boolean
-from biothings.utils.common import timesofar
 import biothings.utils.aws as aws
-from biothings.utils.hub_db import get_cmd, get_last_command, backup, restore
-from biothings.utils.loggers import ShellLogger
-from biothings.utils.web.es import flatten_doc
 from biothings import config
-if hasattr(config, 'logger'):
+from biothings.utils.common import timesofar
+from biothings.utils.dataload import to_boolean
+from biothings.utils.docs import flatten_doc
+from biothings.utils.hub_db import backup, get_cmd, get_last_command, restore
+from biothings.utils.loggers import ShellLogger
+from biothings.utils.redirect_streams import RedirectStdStreams
+
+if hasattr(config, "logger"):
     logging = config.logger
 else:
     import logging
@@ -34,7 +36,7 @@ else:
 # useful variables to bring into hub namespace
 pending = "pending"
 
-HUB_ENV = hasattr(config, "HUB_ENV") and config.HUB_ENV or ""    # default: prod (or "normal")
+HUB_ENV = hasattr(config, "HUB_ENV") and config.HUB_ENV or ""  # default: prod (or "normal")
 VERSIONS = HUB_ENV and "%s-versions" % HUB_ENV or "versions"
 LATEST = HUB_ENV and "%s-latest" % HUB_ENV or "latest"
 
@@ -42,27 +44,51 @@ LATEST = HUB_ENV and "%s-latest" % HUB_ENV or "latest"
 def jsonreadify(cmd):
     newcmd = copy.copy(cmd)
     newcmd.pop("jobs")
+
+    # try to make the data structure of the returned
+    # results for each command is hubdb compatible.
+
+    if "results" in newcmd:
+        results = []
+        for result in newcmd.pop("results"):
+            if isinstance(result, UserDict):
+                results.append(result.data)
+            #
+            # elif isinstance(result, ...):
+            #     pass # add more here
+            #
+            else:  # already compliant
+                results.append(result)
+        newcmd["results"] = results
+
     return newcmd
 
+
 ##############
 # HUB SERVER #
 ##############
 
+
 class AlreadyRunningException(Exception):
     pass
 
+
 class CommandError(Exception):
     pass
 
+
 class NoSuchCommand(Exception):
     pass
 
+
 class CommandNotAllowed(Exception):
     pass
 
+
 class CommandInformation(dict):
     pass
 
+
 class CommandDefinition(dict):
     pass
 
@@ -79,14 +105,15 @@ class HubShell(InteractiveShell):
         self.commands = OrderedDict()
         self.managers = {}
         self.extra_ns = OrderedDict()
-        self.tracked = {}   # command calls kept in history or not
-        self.hidden = {}    # not returned by help()
+        self.tracked = {}  # command calls kept in history or not
+        self.hidden = {}  # not returned by help()
         self.origout = sys.stdout
         self.buf = io.StringIO()
         self.shellog = ShellLogger(name="shell")
         # there should be only one shell instance (todo: singleton)
         self.__class__.cmd = get_cmd()
         self.__class__.set_command_counter()
+        self.last_std_contents = None
         super(HubShell, self).__init__(user_ns=self.extra_ns)
 
     @classmethod
@@ -117,13 +144,17 @@ class HubShell(InteractiveShell):
                         self.tracked[name] = cmd.get("tracked", True)
                         self.hidden[name] = cmd.get("hidden", False) or hidden
                     except KeyError as e:
-                        raise CommandError("Could not register command because missing '%s' in definition: %s" % (e, cmd))
+                        raise CommandError(
+                            "Could not register command because missing '%s' in definition: %s"
+                            % (e, cmd)
+                        )
                 else:
                     self.commands[name] = cmd
                     self.hidden[name] = hidden
                 # update original passed commands to caller knows what's been done there
                 if name not in basic_commands:
                     basic_commands[name] = cmd
+
         # update with ssh server default commands
         register(basic_commands)
         # don't track this calls
@@ -133,16 +164,20 @@ class HubShell(InteractiveShell):
         register({"restore": CommandDefinition(command=restore, track=True)})
         register({"help": CommandDefinition(command=self.help, track=False)})
         register({"commands": CommandDefinition(command=self.command_info, tracked=False)})
-        register({"command": CommandDefinition(
-            command=lambda id, *args, **kwargs: self.command_info(id=id, *args, **kwargs),
-            tracked=False
-        )})
+        register(
+            {
+                "command": CommandDefinition(
+                    command=lambda id, *args, **kwargs: self.command_info(id=id, *args, **kwargs),
+                    tracked=False,
+                )
+            }
+        )
 
         for extra in extra_ns:
             # don't expose extra commands, they're kind of private/advanced
             register(extra, hidden=True)
 
-        #self.extra_ns["cancel"] = self.__class__.cancel
+        # self.extra_ns["cancel"] = self.__class__.cancel
         # for boolean calls
         self.extra_ns["_and"] = _and
         self.extra_ns["partial"] = partial
@@ -159,9 +194,7 @@ class HubShell(InteractiveShell):
         return self.restart(force=force, stop=True)
 
     def restart(self, force=False, stop=False):
-
-        @asyncio.coroutine
-        def do():
+        async def do():
             try:
                 if stop:
                     event = "hub_stop"
@@ -170,20 +203,18 @@ class HubShell(InteractiveShell):
                     event = "hub_restart"
                     msg = "Hub is restarting"
                 logging.critical(
-                    json.dumps({
-                        "type": "alert",
-                        "msg": msg,
-                        "event": event
-                    }),
-                    extra={"event": True}
+                    json.dumps({"type": "alert", "msg": msg, "event": event}),
+                    extra={"event": True},
                 )
                 logging.info("Stopping job manager...")
                 j = self.job_manager.stop(force=force)
+
                 def ok(f):
                     f.result()  # consume
                     logging.error("Job manager stopped")
+
                 j.add_done_callback(ok)
-                yield from j
+                await j
             except Exception as e:
                 logging.error("Error while recycling the process queue: %s", e)
                 raise
@@ -192,6 +223,7 @@ class HubShell(InteractiveShell):
             f.result()  # consume future's result to potentially raise exception
             logging.debug("%s %s", [sys.executable], sys.argv)
             import subprocess
+
             subprocess.Popen([sys.executable] + sys.argv)
             self.job_manager.hub_process.kill()
             sys.exit(0)
@@ -225,7 +257,10 @@ class HubShell(InteractiveShell):
             return cmds
         elif isinstance(func, partial):
             docstr = "\n" + pydoc.render_doc(func.func, title="Hub documentation: %s")
-            docstr += "\nDefined et as a partial, with:\nargs:%s\nkwargs:%s\n" % (repr(func.args), repr(func.keywords))
+            docstr += "\nDefined et as a partial, with:\nargs:%s\nkwargs:%s\n" % (
+                repr(func.args),
+                repr(func.keywords),
+            )
             return docstr
         elif isinstance(func, CompositeCommand):
             docstr = "\nComposite command:\n\n%s\n" % func
@@ -261,7 +296,7 @@ class HubShell(InteractiveShell):
     def extract_command_name(self, cmd):
         try:
             # extract before () (non-callable are never tracked)
-            grps = re.fullmatch("([\w\.]+)(\(.*\))", cmd.strip()).groups()
+            grps = re.fullmatch(r"([\w\.]+)(\(.*\))", cmd.strip()).groups()
             return grps[0]
         except AttributeError:
             raise CommandError("Can't extract command name from '%s'" % repr(cmd))
@@ -291,13 +326,15 @@ class HubShell(InteractiveShell):
             logging.debug("Can't extract command from %s, can't register", cmd)
             return result
         # also, never register non-callable command
-        if not force and (not callable(self.extra_ns.get(cmdname)) or \
-                          self.tracked.get(cmdname, True) == False):
+        if not force and (
+            not callable(self.extra_ns.get(cmdname)) or self.tracked.get(cmdname, True) is False
+        ):
             return result
 
         cmdnum = self.__class__.cmd_cnt
-        cmdinfo = CommandInformation(cmd=cmd, jobs=result, started_at=time.time(),
-                                     id=cmdnum, is_done=False)
+        cmdinfo = CommandInformation(
+            cmd=cmd, jobs=result, started_at=time.time(), id=cmdnum, is_done=False
+        )
         assert cmdnum not in self.__class__.launched_commands
         # register
         self.__class__.launched_commands[cmdnum] = cmdinfo
@@ -305,11 +342,18 @@ class HubShell(InteractiveShell):
         self.__class__.cmd_cnt += 1
 
         # TODO: clean up the following if statement, confirm with Sebastien
-        if type(result) == asyncio.tasks.Task or type(result) == asyncio.tasks._GatheringFuture or \
-                type(result) == asyncio.Future or \
-                type(result) == list and len(result) > 0 and type(result[0]) == asyncio.tasks.Task:
+        if (
+            type(result) == asyncio.tasks.Task
+            or type(result) == asyncio.tasks._GatheringFuture
+            or type(result) == asyncio.Future
+            or type(result) == list
+            and len(result) > 0
+            and type(result[0]) == asyncio.tasks.Task
+        ):
             # it's asyncio related
-            result = type(result) != list and [result] or result   # TODO: cleanup and confirm this line
+            result = (
+                type(result) != list and [result] or result
+            )  # TODO: cleanup and confirm this line
             cmdinfo["jobs"] = result
             return cmdinfo
         else:
@@ -326,7 +370,9 @@ class HubShell(InteractiveShell):
         self.shellog.input(line)
         origline = line  # keep what's been originally entered
         # poor man's singleton...
-        if line in [j["cmd"] for j in self.__class__.launched_commands.values() if not j.get("is_done")]:
+        if line in [
+            j["cmd"] for j in self.__class__.launched_commands.values() if not j.get("is_done")
+        ]:
             raise AlreadyRunningException("Command '%s' is already running\n" % repr(line))
         # is it a hub command, in which case, intercept and run the actual declared cmd
         # IMPORTANT !!! this is where we allow the command or not when secure=True IMPORTANT !!!
@@ -338,15 +384,14 @@ class HubShell(InteractiveShell):
             # command must be alpha only, argument with "," and "=", or no arg at all
             pat = r'^([A-Za-z_]+)\(["\'\w\s=,.-]*\)$'
         else:
-            pat = r'(.*)\(.*\)'  # more permissive
+            pat = r"(.*)\(.*\)"  # more permissive
         m = re.match(pat, line)
         if m:
             cmd = m.groups()[0].strip()
             if secure and cmd not in self.commands:
                 #  match regex rule but not a valid/existing command, discard it
                 raise NoSuchCommand(cmd)
-            if cmd in self.commands and \
-                    isinstance(self.commands[cmd], CompositeCommand):
+            if cmd in self.commands and isinstance(self.commands[cmd], CompositeCommand):
                 line = self.commands[cmd]
         elif line != "" and secure:
             # we have something entered, it doesn't match our regex rule, discard it
@@ -362,7 +407,7 @@ class HubShell(InteractiveShell):
                 # to the partials
                 strcmds = []
                 for one_cmd in chained_cmds:
-                    func, args = re.match("(.*)\((.*)\)", one_cmd).groups()
+                    func, args = re.match(r"(.*)\((.*)\)", one_cmd).groups()
                     if args:
                         strcmds.append("partial(%s,%s)" % (func, args))
                     else:
@@ -370,8 +415,13 @@ class HubShell(InteractiveShell):
                 cmdline = "_and(%s)" % ",".join(strcmds)
             else:
                 raise CommandError("Using '&&' operator required two operands\n")
-        r = self.run_cell(cmdline, store_history=True)
+
+        # r = self.run_cell(cmdline, store_history=True)
         outputs = []
+        with RedirectStdStreams() as redirect_stream:
+            r = self.run_cell(cmdline, store_history=True)
+            self.last_std_contents = redirect_stream.get_std_contents()
+
         if not r.success:
             raise CommandError("%s\n" % repr(r.error_in_exec))
         else:
@@ -407,8 +457,8 @@ class HubShell(InteractiveShell):
 
         return outputs
 
-    #@classmethod
-    #def cancel(klass,jobnum):
+    # @classmethod
+    # def cancel(klass,jobnum):
     #    return klass.launched_commands.get(jobnum)
 
     @classmethod
@@ -417,27 +467,44 @@ class HubShell(InteractiveShell):
             # already process, this current command is now history
             # Note: if we have millions of commands there, it could last quite a while,
             # but IRL we only have a few
-            if info.get("is_done") == True:
+            if info.get("is_done") is True:
                 continue
             # is_done = set([j.done() for j in info["jobs"]]) == set([True])   # TODO: remove this line
             is_done = {j.done() for j in info["jobs"]} == {True}
             has_err = is_done and [True for j in info["jobs"] if j.exception()] or None
-            localoutputs = is_done and ([str(j.exception()) for j in info["jobs"] if j.exception()] or \
-                        [j.result() for j in info["jobs"]]) or None
+            localoutputs = (
+                is_done
+                and (
+                    [str(j.exception()) for j in info["jobs"] if j.exception()]
+                    or [j.result() for j in info["jobs"]]
+                )
+                or None
+            )
             if is_done:
                 cls.launched_commands[num]["is_done"] = True
                 cls.launched_commands[num]["failed"] = has_err and has_err[0] or False
                 cls.launched_commands[num]["results"] = localoutputs
                 cls.launched_commands[num]["finished_at"] = time.time()
-                cls.launched_commands[num]["duration"] = timesofar(t0=cls.launched_commands[num]["started_at"],
-                                                                   t1=cls.launched_commands[num]["finished_at"])
+                cls.launched_commands[num]["duration"] = timesofar(
+                    t0=cls.launched_commands[num]["started_at"],
+                    t1=cls.launched_commands[num]["finished_at"],
+                )
                 cls.save_cmd(num, cls.launched_commands[num])
                 if not has_err and localoutputs and set(map(type, localoutputs)) == {str}:
                     localoutputs = "\n" + "".join(localoutputs)
-                cls.pending_outputs[num] = "[%s] %s {%s} %s: finished %s " % \
-                        (num, has_err and "ERR" or "OK", timesofar(info["started_at"]), info["cmd"], localoutputs)
+                cls.pending_outputs[num] = "[%s] %s {%s} %s: finished %s " % (
+                    num,
+                    has_err and "ERR" or "OK",
+                    timesofar(info["started_at"]),
+                    info["cmd"],
+                    localoutputs,
+                )
             else:
-                cls.pending_outputs[num] = "[%s] RUN {%s} %s" % (num, timesofar(info["started_at"]), info["cmd"])
+                cls.pending_outputs[num] = "[%s] RUN {%s} %s" % (
+                    num,
+                    timesofar(info["started_at"]),
+                    info["cmd"],
+                )
 
     @classmethod
     def command_info(cls, id=None, running=None, failed=None):
@@ -461,7 +528,7 @@ class HubShell(InteractiveShell):
                 # running or done commands (not both)
                 if cmd.get("is_done") == is_done:
                     # done + failed (a failed command is always done btw)
-                    if failed is not None and cmd.get("is_done") == True:
+                    if failed is not None and cmd.get("is_done") is True:
                         if cmd.get("failed") == failed:
                             cmds[_id] = jsonreadify(cmd)
                     else:
@@ -477,7 +544,7 @@ class HubShell(InteractiveShell):
                     continue
             else:
                 # either running or done commands (both)
-                if failed is not None and cmd.get("is_done") == True:
+                if failed is not None and cmd.get("is_done") is True:
                     if cmd.get("failed") == failed:
                         cmds[_id] = jsonreadify(cmd)
                 else:
@@ -495,6 +562,7 @@ class HubShell(InteractiveShell):
 # required arguments, eg.:
 # {"schedule" ; partial(schedule,loop)}
 
+
 def stats(src_dump):
     pass
 
@@ -513,7 +581,7 @@ def template_out(field, confdict):
         => "one_two_three_201908" # assuming we're in August 2019
     """
     # first deal with timestamp
-    pat = re.compile(".*(\$\((.*?)\)).*")
+    pat = re.compile(r".*(\$\((.*?)\)).*")
 
     try:
         m = pat.match(field)
@@ -532,8 +600,9 @@ def template_out(field, confdict):
     return field
 
 
-def publish_data_version(s3_bucket, s3_folder, version_info, update_latest=True,
-                         aws_key=None, aws_secret=None):
+def publish_data_version(
+    s3_bucket, s3_folder, version_info, update_latest=True, aws_key=None, aws_secret=None
+):
     """
     Update remote files:
         - versions.json: add version_info to the JSON list
@@ -552,15 +621,13 @@ def publish_data_version(s3_bucket, s3_folder, version_info, update_latest=True,
     # register version
     versionskey = os.path.join(s3_folder, "%s.json" % VERSIONS)
     try:
-        versions = aws.get_s3_file(versionskey,
-                                   return_what="content",
-                                   aws_key=aws_key,
-                                   aws_secret=aws_secret,
-                                   s3_bucket=s3_bucket)
-        versions = json.loads(versions.decode())  # S3 returns bytes
+        versions = json.loads(
+            aws.get_s3_file_contents(
+                versionskey, aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket
+            ).decode()
+        )
     except (FileNotFoundError, json.JSONDecodeError):
         versions = {"format": "1.0", "versions": []}
-    # if type(version_info) == list:     # remove this line
     if isinstance(version_info, list):
         versions["versions"] = version_info
     else:
@@ -573,48 +640,34 @@ def publish_data_version(s3_bucket, s3_folder, version_info, update_latest=True,
         # order by build_version
         versions["versions"] = sorted(tmp.values(), key=lambda e: e["build_version"])
 
-    aws.send_s3_file(None, versionskey,
-                     content=json.dumps(versions, indent=True),
-                     aws_key=aws_key,
-                     aws_secret=aws_secret,
-                     s3_bucket=s3_bucket,
-                     content_type="application/json",
-                     overwrite=True)
+    aws.send_s3_file(
+        None,
+        versionskey,
+        content=json.dumps(versions, indent=True),
+        aws_key=aws_key,
+        aws_secret=aws_secret,
+        s3_bucket=s3_bucket,
+        content_type="application/json",
+        overwrite=True,
+    )
 
     # update latest
-    # if type(version_info) != list and update_latest:    # TODO: remove this line
     if not isinstance(version_info, list) and update_latest:
         latestkey = os.path.join(s3_folder, "%s.json" % LATEST)
-        key = None
-        try:
-            key = aws.get_s3_file(
-                latestkey,
-                return_what="key",
-                aws_key=aws_key,
-                aws_secret=aws_secret,
-                s3_bucket=s3_bucket
-            )
-        except FileNotFoundError:
-            pass
+        newredir = os.path.join("/", s3_folder, "{}.json".format(version_info["build_version"]))
+        # the consensus is that we will upload the data and have the
+        # redirection, for record-keep purpose
         aws.send_s3_file(
-            None, latestkey,
+            None,
+            latestkey,
             content=json.dumps(version_info["build_version"], indent=True),
             content_type="application/json",
             aws_key=aws_key,
             aws_secret=aws_secret,
             s3_bucket=s3_bucket,
-            overwrite=True
+            overwrite=True,
+            redirect=newredir,
         )
-        if not key:
-            key = aws.get_s3_file(
-                latestkey,
-                return_what="key",
-                aws_key=aws_key,
-                aws_secret=aws_secret,
-                s3_bucket=s3_bucket
-            )
-        newredir = os.path.join("/", s3_folder, "{}.json".format(version_info["build_version"]))
-        key.set_redirect(newredir)
 
 
 def _and(*funcs):
@@ -636,10 +689,12 @@ def _and(*funcs):
         raise CommandError("First command didn't return a future, can't chain commands")
     all_res.append(fut1)
     # err = None
+
     def do(f, cb):
         f.result()  # consume exception if any
         if cb:
             all_res.extend(_and(cb, *funcs))
+
     if len(funcs) > 1:
         func2 = funcs[1]
         if len(funcs) > 2:
@@ -656,21 +711,22 @@ class CompositeCommand(str):
     a new command made of other commands. Useful to define
     shortcuts when typing commands in hub console.
     """
+
     def __init__(self, cmd):
         self.cmd = cmd
 
     def __str__(self):
         return "<CompositeCommand: '%s'>" % self.cmd
 
+
 ############
 # RELOADER #
 ############
 
+
 def exclude_from_reloader(path):
     # exlucde cached, git and hidden files
-    return path.endswith("__pycache__") or \
-           ".git" in path or \
-           os.path.basename(path).startswith(".")
+    return path.endswith("__pycache__") or ".git" in path or os.path.basename(path).startswith(".")
 
 
 class BaseHubReloader(object):
@@ -694,91 +750,100 @@ class BaseHubReloader(object):
         raise NotImplementedError("Implement me in a sub-class")
 
 
-class PyInotifyHubReloader(BaseHubReloader):
-    """Based on pyinotify events"""
-
-    # inner class to hide pyinotify in case not available
+try:
     import pyinotify
-    class ReloadListener(pyinotify.ProcessEvent):
-    
-        def my_init(self, watcher_manager, reload_func):
-            self.reload_func = reload_func
-            self.watcher_manager = watcher_manager
-    
-        def process_default(self, event):
-            pyinotify = sys.modules["pyinotify"]
-            if exclude_from_reloader(event.pathname):
-                return
-            if event.dir:
-                if event.mask & pyinotify.IN_CREATE:
-                    # add to watcher. no need to check if already watched, manager knows
-                    # how to deal with that
-                    logging.info("Add '%s' to watcher", event.pathname)
-                    self.watcher_manager.add_watch(event.pathname, self.notifier.mask, rec=1)
-                elif event.mask & pyinotify.IN_DELETE:
-                    logging.info("Remove '%s' from watcher", event.pathname)
-                    # watcher knows when directory is deleted (file descriptor become invalid),
-                    # so no need to do it manually
-            logging.error("Need to reload manager because of %s", event)
-            self.reload_func()
-    
 
-    def __init__(self, paths, reload_func, wait=5, mask=None):
-        pyinotify = sys.modules["pyinotify"] # get it from sys.modules or we'd need another "import"
-                                             # just sure why...
-        if isinstance(paths, str):
-            paths = [paths]
-        paths = set(paths)  # get rid of duplicated, just in case
-        self.mask = mask or pyinotify.IN_CREATE|pyinotify.IN_DELETE|pyinotify.IN_CLOSE_WRITE
-        self.reload_func = reload_func
-        # only listen to these events. Note: directory detection is done via a flag so
-        # no need to use IS_DIR
-        self.watcher_manager = pyinotify.WatchManager(exclude_filter=exclude_from_reloader)
-        self.paths = []  # cleaned
-        for path in paths:
-            if not os.path.isabs(path):
-                path = os.path.abspath(path)
-            self.paths.append(path)
-            self.watcher_manager.add_watch(path, self.mask, rec=True, exclude_filter=exclude_from_reloader)   # recursive
+    class PyInotifyHubReloader(BaseHubReloader):
+        """Based on pyinotify events"""
 
-        self.listener = self.__class__.ReloadListener(
-                watcher_manager=self.watcher_manager,
-                reload_func=self.reload_func)
-        self.notifier = pyinotify.Notifier(
-                self.watcher_manager,
-                default_proc_fun=self.listener
-                )
-        # propagate notifier so notifer itself can be reloaded (when new directory/source)
-        self.listener.notifier = self
-        self.wait = wait
-        self.do_monitor = False
+        # inner class to hide pyinotify in case not available
+        class ReloadListener(pyinotify.ProcessEvent):
+            def my_init(self, watcher_manager, reload_func):
+                self.reload_func = reload_func
+                self.watcher_manager = watcher_manager
 
-    def monitor(self):
+            def process_default(self, event):
+                pyinotify = sys.modules["pyinotify"]
+                if exclude_from_reloader(event.pathname):
+                    return
+                if event.dir:
+                    if event.mask & pyinotify.IN_CREATE:
+                        # add to watcher. no need to check if already watched, manager knows
+                        # how to deal with that
+                        logging.info("Add '%s' to watcher", event.pathname)
+                        self.watcher_manager.add_watch(event.pathname, self.notifier.mask, rec=1)
+                    elif event.mask & pyinotify.IN_DELETE:
+                        logging.info("Remove '%s' from watcher", event.pathname)
+                        # watcher knows when directory is deleted (file descriptor become invalid),
+                        # so no need to do it manually
+                logging.error("Need to reload manager because of %s", event)
+                self.reload_func()
 
-        @asyncio.coroutine
-        def do():
-            logging.info(
-                "Monitoring source code in, %s:\n%s",
-                repr(self.paths),
-                pformat(self.watched_files())
+        def __init__(self, paths, reload_func, wait=5, mask=None):
+            pyinotify = sys.modules[
+                "pyinotify"
+            ]  # get it from sys.modules or we'd need another "import"
+            # just sure why...
+            if isinstance(paths, str):
+                paths = [paths]
+            paths = set(paths)  # get rid of duplicated, just in case
+            self.mask = (
+                mask or pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_CLOSE_WRITE
             )
-            while self.do_monitor:
-                try:
-                    yield from asyncio.sleep(self.wait)
-                    # this reads events from OS
-                    if self.notifier.check_events(0.1):
-                        self.notifier.read_events()
-                    # Listener gets called there if event avail
-                    self.notifier.process_events()
-                except KeyboardInterrupt:
-                    logging.warning("Stop monitoring code")
-                    break
+            self.reload_func = reload_func
+            # only listen to these events. Note: directory detection is done via a flag so
+            # no need to use IS_DIR
+            self.watcher_manager = pyinotify.WatchManager(exclude_filter=exclude_from_reloader)
+            self.paths = []  # cleaned
+            for path in paths:
+                if not os.path.isabs(path):
+                    path = os.path.abspath(path)
+                self.paths.append(path)
+                self.watcher_manager.add_watch(
+                    path, self.mask, rec=True, exclude_filter=exclude_from_reloader
+                )  # recursive
 
-        self.do_monitor = True
-        return asyncio.ensure_future(do())
+            self.listener = self.__class__.ReloadListener(
+                watcher_manager=self.watcher_manager, reload_func=self.reload_func
+            )
+            self.notifier = pyinotify.Notifier(
+                self.watcher_manager, default_proc_fun=self.listener
+            )
+            # propagate notifier so notifer itself can be reloaded (when new directory/source)
+            self.listener.notifier = self
+            self.wait = wait
+            self.do_monitor = False
 
-    def watched_files(self):
-        return [v.path for v in self.watcher_manager.watches.values()]
+        def monitor(self):
+            async def do():
+                logging.info(
+                    "Monitoring source code in, %s:\n%s",
+                    repr(self.paths),
+                    pformat(self.watched_files()),
+                )
+                while self.do_monitor:
+                    try:
+                        await asyncio.sleep(self.wait)
+                        # this reads events from OS
+                        if self.notifier.check_events(0.1):
+                            self.notifier.read_events()
+                        # Listener gets called there if event avail
+                        self.notifier.process_events()
+                    except KeyboardInterrupt:
+                        logging.warning("Stop monitoring code")
+                        break
+
+            self.do_monitor = True
+            return asyncio.ensure_future(do())
+
+        def watched_files(self):
+            return [v.path for v in self.watcher_manager.watches.values()]
+
+except ImportError:
+
+    class PyInotifyHubReloader:
+        def __new__(cls, *args, **kwargs):
+            raise RuntimeError("Can't create PyInotifyHubReloader instance" "without inotify")
 
 
 class TornadoAutoReloadHubReloader(BaseHubReloader):
@@ -814,16 +879,13 @@ class TornadoAutoReloadHubReloader(BaseHubReloader):
 
     def monitor(self):
         logging.info(
-            "Monitoring source code in, %s:\n%s",
-            repr(self.paths),
-            pformat(self.watched_files())
+            "Monitoring source code in, %s:\n%s", repr(self.paths), pformat(self.watched_files())
         )
-        self.mod.start(self.wait*1000) # millis
+        self.mod.start(self.wait * 1000)  # millis
 
     def watched_files(self):
         return self.mod._watched_files
-        #return [d for d in self.mod._watched_files if os.path.isdir(d)]
-
+        # return [d for d in self.mod._watched_files if os.path.isdir(d)]
 
 
 def get_hub_reloader(*args, **kwargs):
@@ -833,13 +895,15 @@ def get_hub_reloader(*args, **kwargs):
     """
     if getattr(config, "USE_RELOADER", False) and config.USE_RELOADER:
         try:
-            import pyinotify
+            import pyinotify  # noqa
+
             logging.info("Using Hub reloader based on pyinotify")
-            return PyInotifyHubReloader(*args,**kwargs)
+            return PyInotifyHubReloader(*args, **kwargs)
         except ImportError:
-            import tornado.autoreload
+            import tornado.autoreload  # noqa
+
             logging.info("Using Hub reloader based on tornado.autoreload")
-            return TornadoAutoReloadHubReloader(*args,**kwargs)
+            return TornadoAutoReloadHubReloader(*args, **kwargs)
     else:
         logging.info("USE_RELOADER not set (or False), won't monitor for changes")
         return None
